import dataclasses
import weakref
from typing import Optional
from javaobj import JavaObject
from .BackgroundImage import BackgroundImage
from .ModelBase import ModelBase
from ..AssetManager import AssetManager

_level_cache = weakref.WeakValueDictionary()

@dataclasses.dataclass(unsafe_hash=True)
class Level(ModelBase):
    identifier: str
    name: str
    elevation: float
    floor_thickness: float
    height: float
    background_image: Optional[BackgroundImage]
    is_visible: bool
    is_viewable: bool
    elevation_index: int

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Level':
        identifier = o.id
        if identifier in _level_cache:
            return _level_cache[identifier]

        level = cls(
            identifier=identifier,
            name=o.name,
            elevation=o.elevation,
            floor_thickness=o.floorThickness,
            height=o.height,
            background_image=BackgroundImage.from_javaobj(o.backgroundImage, asset_manager) if o.backgroundImage else None,
            is_visible=o.visible,
            is_viewable=o.viewable,
            elevation_index=o.elevationIndex
        )

        _level_cache[identifier] = level

        return level

    @classmethod
    def from_identifier(cls, identifier: str) -> 'Level':
        wall = _level_cache.get(identifier)
        if not wall:
            raise ValueError('Identifier not found in cache')
        return wall

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Level':
        identifier = data.get('@id')
        if not identifier:
            raise ValueError('@id is required')
        background_image = data.get('backgroundImage')
        if identifier in _level_cache:
            return _level_cache[identifier]

        level = cls(
            identifier=identifier,
            name=cls.required_str(data.get('@name')),
            elevation=cls.required_float(data.get('@elevation')),
            floor_thickness=cls.required_float(data.get('@floorThickness')),
            height=cls.required_float(data.get('@height')),
            background_image=BackgroundImage.from_xml_dict(background_image, asset_manager) if background_image else None,
            is_visible=data.get('@visible', 'true') == 'true',
            is_viewable=data.get('@viewable', 'true') == 'true',
            elevation_index=cls.required_int(data.get('@elevationIndex'))
        )

        _level_cache[identifier] = level

        return level



from abc import ABC, abstractmethod
from typing import Optional
from javaobj import JavaObject
from ..AssetManager import AssetManager


class ModelBase(ABC):

    @staticmethod
    def required_float(value: Optional[str]) -> float:
        if not value:
            raise ValueError('Value is required')
        return float(value)

    @staticmethod
    def required_int(value: Optional[str]) -> int:
        if not value:
            raise ValueError('Value is required')
        return int(value)

    @staticmethod
    def required_str(value: Optional[str]) -> str:
        if not value:
            raise ValueError('Value is required')
        return value

    @classmethod
    @abstractmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'ModelBase':
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'ModelBase':
        raise NotImplementedError



import dataclasses
from typing import Optional
from javaobj import JavaObject
from .HomeTexture import HomeTexture
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


@dataclasses.dataclass
class HomeMaterial(ModelBase):
    name: str
    color: Optional[int]
    texture: Optional[HomeTexture]
    shininess: Optional[float]

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'HomeMaterial':
        return cls(
            name=o.name,
            color=o.color,
            texture=HomeTexture.from_javaobj(o.texture, asset_manager) if o.texture else None,
            shininess=o.shininess,
        )



import dataclasses
from functools import cached_property
from typing import List, TypeVar, Tuple, Optional
from javaobj import JavaObject

from .HasLevel import HasLevel
from .Level import Level
from .BackgroundImage import BackgroundImage
from .HomePieceOfFurniture import HomePieceOfFurniture
from .HomeDoorOrWindow import HomeDoorOrWindow
from .Polyline import Polyline
from .Room import Room
from .Wall import Wall
from .Renderable import Renderable
from .DimensionLine import DimensionLine
from .Label import Label
from .HomeEnvironment import HomeEnvironment
from .Compass import Compass
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


L = TypeVar('L', bound=HasLevel)

@dataclasses.dataclass
class Home(ModelBase):
    levels: List[Level]  # Implemented
    furniture: List[HomePieceOfFurniture]  # Implemented
    rooms: List[Room]  # Implemented
    walls: List[Wall]  # Implemented
    dimension_lines: List[DimensionLine]  # Implemented
    labels: List[Label] # Implemented
    polylines: List[Polyline]  # Implemented
    wall_height: float
    name: str
    environment: HomeEnvironment  # Ignored, not needed for SVG
    compass: Compass  # Not really needed?
    version: int
    background_image: Optional[BackgroundImage] = None  # Implemented

    @cached_property
    def selectable_viewable_items(self) -> List[Renderable]:
        renderable = self.walls + self.rooms + self.dimension_lines + self.labels + self.polylines
        viewable_items: List[Renderable] = [item for item in renderable if not item.level or item.level.is_visible]

        for furniture in self.furniture:
            if furniture.is_visible and (not furniture.level or furniture.level.is_visible):
                viewable_items.append(furniture)

                if isinstance(furniture, HomeDoorOrWindow):
                    for sash in furniture.sashes:
                        viewable_items.append(sash)

        return viewable_items

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Home':
        walls = [Wall.from_javaobj(wo, asset_manager) for wo in o.walls]
        if hasattr(o, 'furnitureWithDoorsAndWindows'):
            furniture_obj = o.furnitureWithDoorsAndWindows
        else:
            furniture_obj = o.furniture

        furniture = []
        for hpofo in furniture_obj:
            if hpofo.classdesc.name == 'com.eteks.sweethome3d.model.HomeDoorOrWindow':
                furniture.append(HomeDoorOrWindow.from_javaobj(hpofo, asset_manager))
            else:
                furniture.append(HomePieceOfFurniture.from_javaobj(hpofo, asset_manager))

        return cls(
            levels=[Level.from_javaobj(lo, asset_manager) for lo in o.levels],
            furniture=furniture,
            rooms=[Room.from_javaobj(ro, asset_manager) for ro in o.rooms],
            walls=walls,
            dimension_lines=[DimensionLine.from_javaobj(dlo, asset_manager) for dlo in o.dimensionLines],
            polylines=[Polyline.from_javaobj(po, asset_manager) for po in o.polylines],
            labels=[Label.from_javaobj(lo, asset_manager) for lo in o.labels],
            wall_height=o.wallHeight,
            name=o.name,
            background_image=BackgroundImage.from_javaobj(o.backgroundImage, asset_manager) if o.backgroundImage else None,
            environment=HomeEnvironment.from_javaobj(o.environment, asset_manager),
            compass=Compass.from_javaobj(o.compass, asset_manager),
            version=o.version
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Home':
        background_image = data.get('backgroundImage')
        dimension_line = data.get('dimensionLine', [])
        compass = data.get('compass')

        if not compass:
            raise ValueError('compass is not provided')

        if not isinstance(dimension_line, list):
            dimension_line = [dimension_line]
        label = data.get('label', [])
        if not isinstance(label, list):
            label = [label]

        wall = data.get('wall', [])
        if not isinstance(wall, list):
            wall = [wall]

        level = data.get('level', [])
        if not isinstance(level, list):
            level = [level]

        # Levels needs to be processed first to fill in cache
        levels = [Level.from_xml_dict(le, asset_manager) for le in level]

        door_or_window = data.get('doorOrWindow',[])
        if not isinstance(door_or_window, list):
            door_or_window = [door_or_window]
        piece_of_furniture = data.get('pieceOfFurniture',[])
        if not isinstance(piece_of_furniture, list):
            piece_of_furniture = [piece_of_furniture]

        furniture = [HomeDoorOrWindow.from_xml_dict(dow, asset_manager) for dow in door_or_window]
        furniture += [HomePieceOfFurniture.from_xml_dict(hpofo, asset_manager) for hpofo in piece_of_furniture]

        # Stitch together walls
        walls: List[Tuple[Wall, str, str]] = []
        for wo in wall:
            wall_at_start_identifier = wo.get('@wallAtStart')
            wall_at_end_identifier = wo.get('@wallAtEnd')
            walls.append((Wall.from_xml_dict(wo, asset_manager), wall_at_start_identifier, wall_at_end_identifier))

        # Loop over created walls and set wall_at_start/end correctly
        for wall, wall_at_start_identifier, wall_at_end_identifier in walls:
            if wall_at_start_identifier and not wall.wall_at_start:
                wall.wall_at_start = Wall.from_identifier(wall_at_start_identifier)

            if wall_at_end_identifier and not wall.wall_at_end:
                wall.wall_at_end = Wall.from_identifier(wall_at_end_identifier)

        return cls(
            levels=levels,
            labels=[Label.from_xml_dict(lo, asset_manager) for lo in label],
            furniture=furniture,
            rooms=[Room.from_xml_dict(ro, asset_manager) for ro in data.get('room')],
            walls=[w[0] for w in walls],
            dimension_lines=[DimensionLine.from_xml_dict(dl, asset_manager) for dl in dimension_line],
            polylines=[Polyline.from_xml_dict(po, asset_manager) for po in data.get('polyline')],
            wall_height=cls.required_float(data.get('@wallHeight')),
            name=cls.required_str(data.get('@name')),
            background_image=BackgroundImage.from_xml_dict(data.get('backgroundImage'), asset_manager) if background_image else None,
            environment=HomeEnvironment.from_xml_dict(data.get('environment'), asset_manager),
            compass=Compass.from_xml_dict(compass, asset_manager),
            version=cls.required_int(data.get('@version')),
        )



import zipfile
import base64
from pathlib import Path
from typing import Dict, Generator, Optional
from packaging.version import Version
from .model.Content import Content
from .model.ContentDigest import ContentDigest
from .enums.DigestHashEnum import DigestHashEnum

class AssetManager:
    _resources_path: Path
    _assets: Dict[str, Content]
    _textures: Dict[str, Path]
    _patterns: Dict[str, Path]

    def __init__(self, resources_path: Path):
        self._resources_path = resources_path
        self._assets = {}
        self._textures = {}
        self._patterns = {}

        self._preload_textures()
        self._preload_patterns()

    def _preload_textures(self) -> None:
        textures_path = self._resources_path.joinpath('textures/')
        for texture in textures_path.glob('*.jpg'):
            self._textures[texture.stem] = texture

    def _preload_patterns(self) -> None:
        textures_path = self._resources_path.joinpath('patterns/')
        for pattern in textures_path.glob('*.png'):
            self._patterns[pattern.stem] = pattern

    def _parse_content_digests(self, content_digests: bytes) -> Generator[ContentDigest, None, None]:
        content = content_digests.decode()
        header = 'ContentDigests-Version: '
        # File format check

        hash_to_enum = {
            'sha-1': DigestHashEnum.SHA_1
        }

        lines = content.split('\n')
        expecting_header = True
        current_block: Dict[str, str] = {}
        for line in lines:
            if expecting_header:
                if not line.startswith(header):
                    raise ValueError('Unknown file format')
                version = Version(line.replace(header, ''))
                if version < Version('1.0'):
                    raise ValueError('Unknown file format version {}'.format(version))
                expecting_header = False
            else:
                # Parse content
                if not line:
                    if current_block:
                        name = current_block.get('name')
                        digest = current_block.get('digest')
                        digest_hash = current_block.get('digest_hash')
                        if not name or not digest or not digest_hash:
                            raise ValueError
                        yield ContentDigest(
                            name=name,
                            digest_hash=hash_to_enum.get(digest_hash, DigestHashEnum.SHA_1),
                            digest=base64.b64decode(digest)
                        )
                        current_block = {}
                    continue

                key_raw, value = line.split(':')
                key = key_raw.strip().lower()
                if key.endswith('-digest'):
                    hash_string, digest = key.rsplit('-', 1)
                    current_block['digest'] = value.strip()
                    current_block['digest_hash'] = hash_string
                else:
                    current_block[key] = value.strip()

    def load_assets(self, zip_file: zipfile.ZipFile) -> None:
        for content_digest in self._parse_content_digests(zip_file.read('ContentDigests')):
            self._assets[content_digest.name] = Content.from_content_digest(zip_file, content_digest)

    def get_asset(self, asset_name: str) -> Optional[Content]:
        return self._assets.get(asset_name)

    def get_asset_by_uri(self, uri: str) -> Optional[Content]:
        _, asset_name = uri.split('/')
        return self._assets.get(asset_name)

    def get_texture(self, texture_name: str) -> Content:
        found = self._textures.get(texture_name)
        if not found:
            raise ValueError('Requested texture {} was not found'.format(texture_name))
        return Content.from_path(found)

    def get_pattern(self, pattern_name: str) -> Content:
        found = self._patterns.get(pattern_name)
        if not found:
            raise ValueError('Requested pattern {} was not found'.format(pattern_name))
        return Content.from_path(found)



import enum
from importlib.resources import files
from pathlib import Path
from types import TracebackType
from typing import Optional,  Union, Type
import zipfile
import xmltodict
import javaobj

from sh3d.AssetManager import AssetManager
from sh3d.model.Home import Home

@enum.unique
class HomeSource(enum.Enum):
    JAVAOBJ = 'JAVAOBJ'
    XML = 'XML'

class FileLoader:
    home_source: HomeSource
    def __init__(
            self,
            sh3d_path: Union[Path, str],
            resources_path: Optional[Union[Path, str]] = None,
            home_source: HomeSource = HomeSource.JAVAOBJ
    ):
        if isinstance(sh3d_path, str):
            sh3d_path = Path(sh3d_path)

        if isinstance(resources_path, str):
            resources_path = Path(resources_path)

        if not resources_path:
            # Load package provided resources
            resources_path = Path(str(files("sh3d").joinpath("resources")))

        self.asset_manager = AssetManager(resources_path)
        self.zip_file = zipfile.ZipFile(sh3d_path, 'r')  # pylint: disable=consider-using-with
        self.home_source = home_source

    @property
    def home(self) -> Home:
        self.asset_manager.load_assets(self.zip_file)
        return self._load_home()

    def _load_home(self) -> Home:
        if self.home_source == HomeSource.XML:
            data = xmltodict.parse(self.zip_file.read('Home.xml'))
            return Home.from_xml_dict(data.get('home'), self.asset_manager)

        data = javaobj.loads(self.zip_file.read('Home'))
        return Home.from_javaobj(data, self.asset_manager)

    def __enter__(self) -> 'FileLoader':
        return self

    def close(self) -> None:
        self.zip_file.close()

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self.close()



from functools import cached_property
from typing import Tuple
import dataclasses

@dataclasses.dataclass
class Color:
    argb: int

    @cached_property
    def rgba(self) -> Tuple[int, int, int, int]:
        a = (self.argb >> 24) & 0xFF
        r = (self.argb >> 16) & 0xFF
        g = (self.argb >> 8) & 0xFF
        b = self.argb & 0xFF
        return r, g, b, a

    @cached_property
    def alpha_float(self) -> float:
        return max(0.0, min(self.rgba[3] / 255.0, 1.0))

    @cached_property
    def hexa(self) -> str:
        r, g, b, a = self.rgba
        return f'#{r:02X}{g:02X}{b:02X}{a:02X}'

    @cached_property
    def hex(self) -> str:
        r, g, b, _a = self.rgba
        return f'#{r:02X}{g:02X}{b:02X}'

    @classmethod
    def from_argb_interger(cls, argb: int) -> 'Color':
        return cls(
            argb=argb
        )

    @classmethod
    def from_hexa(cls, hexa: str) -> 'Color':
        return cls(
            argb=int.from_bytes(bytes.fromhex(hexa))
        )

    @classmethod
    def from_rgba(cls, r: int, g: int, b: int, a: int) -> 'Color':
        return cls(
            argb=((a & 0xFF) << 24) | ((r & 0xFF) << 16) | ((g & 0xFF) << 8) | (b & 0xFF)
        )



from dataclasses import dataclass
from functools import cached_property
from typing import Optional, List
from javaobj import JavaObject
from shapely.geometry.polygon import Polygon

from .HasLevel import HasLevel
from .Renderable import Renderable, PointType
from .TextStyle import TextStyle, Alignment
from .HomeTexture import HomeTexture
from .Level import Level
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


DEFAULT_ROOM_TEXT_STYLE = TextStyle(
    font_size=24,
    bold=False,
    italic=False,
    alignment=Alignment.CENTER
)

@dataclass
class Room(ModelBase, Renderable, HasLevel):
    identifier: str
    name_x_offset: float
    name_y_offset: float
    is_area_visible: bool
    area_x_offset: float
    area_y_offset: float
    name_angle: float
    area_angle: float
    floor_visible: bool
    floor_shininess: float
    ceiling_visible: bool
    ceiling_shininess: float
    name_style: TextStyle
    area_style: TextStyle
    points: List[PointType]
    name: Optional[str] = None
    floor_color: Optional[int] = None
    floor_texture: Optional[HomeTexture] = None
    ceiling_color: Optional[int] = None
    ceiling_texture: Optional[HomeTexture] = None
    level: Optional[Level] = None

    @cached_property
    def x_center(self) -> float:
        x_coords = [p[0] for p in self.points]
        x_min = min(x_coords)
        x_max = max(x_coords)
        return (x_min + x_max) / 2

    @cached_property
    def y_center(self) -> float:
        y_coords = [p[1] for p in self.points]
        y_min = min(y_coords)
        y_max = max(y_coords)
        return (y_min + y_max) / 2


    @cached_property
    def area(self) -> float:
        return self.geometry.area

    @cached_property
    def geometry(self) -> Polygon:
        return Polygon(self.points)

    def is_at_level(self, level: Level) -> bool:
        if self.level == level:
            return True

        if not self.level or not level:
            return False

        return self.level.elevation == level.elevation and self.level.elevation_index < level.elevation_index

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Room':
        return cls(
            identifier=o.id,
            name=o.name.strip() if o.name else None,
            name_x_offset=o.nameXOffset,
            name_y_offset=o.nameYOffset,
            name_style=TextStyle.from_javaobj(o.nameStyle, asset_manager) if o.nameStyle else DEFAULT_ROOM_TEXT_STYLE,
            name_angle=o.nameAngle,
            points=o.points,
            is_area_visible=o.areaVisible,
            area_x_offset=o.areaXOffset,
            area_y_offset=o.areaYOffset,
            area_style=TextStyle.from_javaobj(o.areaStyle, asset_manager) if o.areaStyle else DEFAULT_ROOM_TEXT_STYLE,
            area_angle=o.areaAngle,
            floor_visible=o.floorVisible,
            floor_color=o.floorColor,
            floor_texture=HomeTexture.from_javaobj(o.floorTexture, asset_manager) if o.floorTexture else None,
            floor_shininess=o.floorShininess,
            ceiling_visible=o.ceilingVisible,
            ceiling_color=o.ceilingColor,
            ceiling_texture=HomeTexture.from_javaobj(o.ceilingTexture, asset_manager) if o.ceilingTexture else None,
            ceiling_shininess=o.ceilingShininess,
            level=Level.from_javaobj(o.level, asset_manager) if o.level else None
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Room':
        level_identifier = data.get('@level')
        points = data.get('point')
        return cls(
            identifier=cls.required_str(data.get('@id')),
            name=data.get('@name'),
            name_x_offset=0.0,
            name_y_offset=-40.0,
            name_style=DEFAULT_ROOM_TEXT_STYLE,
            name_angle=0.0,
            points=[[float(point.get('@x')), float(point.get('@y'))] for point in points],
            is_area_visible=data.get('@areaVisible') == 'true',
            area_x_offset=0.0,
            area_y_offset=0.0,
            area_style=DEFAULT_ROOM_TEXT_STYLE,
            area_angle=0.0,
            floor_visible=True,
            floor_color=0,
            floor_texture=None,
            floor_shininess=0.0,
            ceiling_visible=False,
            ceiling_color=0,
            ceiling_texture=None,
            ceiling_shininess=0.0,
            level=Level.from_identifier(level_identifier) if level_identifier else None
        )



import dataclasses
import enum

from javaobj import JavaObject
from .ModelBase import ModelBase
from ..AssetManager import AssetManager

@enum.unique
class Alignment(enum.Enum):
    LEFT = 'LEFT'
    CENTER = 'CENTER'
    RIGHT = 'RIGHT'

@dataclasses.dataclass
class TextStyle(ModelBase):
    font_size: float
    bold: bool
    italic: bool
    alignment: Alignment

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'TextStyle':
        return cls(
            font_size=o.fontSize,
            bold=o.bold,
            italic=o.italic,
            alignment=Alignment(o.alignment.constant) if o.alignment else Alignment.CENTER
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'ModelBase':
        return cls(
            font_size=cls.required_float(data.get('@fontSize')),
            bold=False,
            italic=False,
            alignment=Alignment.CENTER
        )



import dataclasses
import math
from functools import cached_property
from typing import Union, TYPE_CHECKING, List
from javaobj import JavaObject
from shapely.geometry.linestring import LineString
from shapely.geometry.polygon import Polygon
from shapely.affinity import rotate, translate

from sh3d.AssetManager import AssetManager
from sh3d.geometry import arc, ArcTypeEnum
from .Renderable import Renderable, PointType
from .ModelBase import ModelBase

if TYPE_CHECKING:
    from .HomeDoorOrWindow import HomeDoorOrWindow


@dataclasses.dataclass
class Sash(Renderable):
    door_or_window: 'HomeDoorOrWindow'
    x_axis: float
    y_axis: float
    width: float
    start_angle: float
    end_angle: float

    @cached_property
    def geometry(self) -> Union[LineString, Polygon]:
        model_mirrored_sign = 1.0 if self.door_or_window.is_model_mirrored else -1.0
        x_axis = model_mirrored_sign * self.x_axis * self.door_or_window.width
        y_axis = self.y_axis * self.door_or_window.depth
        sash_width = self.width * self.door_or_window.width
        start_angle = math.degrees(self.start_angle)
        if self.door_or_window.is_model_mirrored:
            start_angle = 180 - start_angle

        extent_angle = model_mirrored_sign * math.degrees(self.end_angle - self.start_angle)
        sash_arc = arc(x_axis - sash_width, y_axis - sash_width, 2 * sash_width, 2 * sash_width, start_angle, extent_angle, ArcTypeEnum.PIE)
        arc_translated = translate(
            sash_arc,
            xoff=self.door_or_window.x - (model_mirrored_sign * -(self.door_or_window.width / 2)) - x_axis,
            yoff=self.door_or_window.y - (self.door_or_window.depth / 2)
        )
        return rotate(
            arc_translated,
            angle=self.door_or_window.angle,
            origin=(self.door_or_window.x, self.door_or_window.y),
            use_radians=True
        )

    @property
    def points(self) -> List[PointType]:
        return list(self.geometry.exterior.coords)

    @points.setter
    def points(self, points: List[PointType]) -> None:
        raise NotImplementedError

    @classmethod
    def from_javaobj(cls, door_or_window: 'HomeDoorOrWindow', o: JavaObject, _asset_manager: AssetManager) -> 'Sash':
        return cls(
            door_or_window=door_or_window,
            x_axis=o.xAxis,
            y_axis=o.yAxis,
            width=o.width,
            start_angle=o.startAngle,
            end_angle=o.endAngle,
        )

    @classmethod
    def from_xml_dict(cls, door_or_window: 'HomeDoorOrWindow', data: dict, _asset_manager: AssetManager) -> 'Sash':
        return cls(
            door_or_window=door_or_window,
            x_axis=ModelBase.required_float(data.get('@xAxis')),
            y_axis=ModelBase.required_float(data.get('@yAxis')),
            width=ModelBase.required_float(data.get('@width')),
            start_angle=ModelBase.required_float(data.get('@startAngle')),
            end_angle=ModelBase.required_float(data.get('@endAngle'))
        )



import dataclasses

from javaobj import JavaObject

from sh3d.AssetManager import AssetManager
from sh3d.model.HomeTexture import HomeTexture
from sh3d.model.ModelBase import ModelBase

@dataclasses.dataclass
class BaseBoard(ModelBase):
    thickness: float
    height: float
    color: int
    texture: HomeTexture

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'ModelBase':
        raise NotImplementedError

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'ModelBase':
        raise NotImplementedError



import dataclasses
import io
import math
from functools import cached_property
from PIL import Image
from PIL.ImageFile import ImageFile
from javaobj import JavaObject
from .Content import Content
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


@dataclasses.dataclass
class BackgroundImage(ModelBase):
    image: Content
    scale_distance: float
    scale_distance_x_start: float
    scale_distance_y_start: float
    scale_distance_x_end: float
    scale_distance_y_end: float
    x_origin: float
    y_origin: float
    invisible: bool

    @cached_property
    def image_info(self) -> ImageFile:
        return Image.open(io.BytesIO(self.image.data))

    @cached_property
    def scale(self) -> float:
        dx = self.scale_distance_x_end - self.scale_distance_x_start
        dy = self.scale_distance_y_end - self.scale_distance_y_start
        pixel_distance = math.hypot(dx, dy)

        # Calculate the scale factor
        return self.scale_distance / pixel_distance


    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'BackgroundImage':
        image = asset_manager.get_asset_by_uri(o.image.url.file)
        if not image:
            raise ValueError('Image was not found in data')
        return cls(
            image=image,
            scale_distance=o.scaleDistance,
            scale_distance_x_start=o.scaleDistanceXStart,
            scale_distance_y_start=o.scaleDistanceYStart,
            scale_distance_x_end=o.scaleDistanceXEnd,
            scale_distance_y_end=o.scaleDistanceYEnd,
            x_origin=o.xOrigin,
            y_origin=o.yOrigin,
            invisible=o.invisible,
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'BackgroundImage':
        image = data.get('@image')
        if not image:
            raise ValueError('image is not provided')

        image_content = asset_manager.get_asset(image)
        if not image_content:
            raise ValueError('Content was not found')

        return cls(
            image=image_content,
            scale_distance=cls.required_float(data.get('@scaleDistance')),
            scale_distance_x_start=cls.required_float(data.get('@scaleDistanceXStart')),
            scale_distance_y_start=cls.required_float(data.get('@scaleDistanceYStart')),
            scale_distance_x_end=cls.required_float(data.get('@scaleDistanceXEnd')),
            scale_distance_y_end=cls.required_float(data.get('@scaleDistanceYEnd')),
            x_origin=cls.required_float(data.get('@xOrigin')),
            y_origin=cls.required_float(data.get('@yOrigin')),
            invisible=False,
        )



import dataclasses
from typing import Optional
from zoneinfo import ZoneInfo
from javaobj import JavaObject
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


@dataclasses.dataclass
class Compass(ModelBase):
    x: float
    y: float
    diameter: float
    visible: bool
    north_direction: float
    latitude: float
    longitude: float
    time_zone: Optional[ZoneInfo]

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Compass':
        return cls(
            x=o.x,
            y=o.y,
            diameter=o.diameter,
            visible=o.visible,
            north_direction=o.northDirection,
            latitude=o.latitude,
            longitude=o.longitude,
            time_zone=ZoneInfo(o.timeZone.ID)
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Compass':
        time_zone = data.get('@timeZone')
        if not time_zone:
            raise ValueError('TimeZone is required')

        return cls(
            x=cls.required_float(data.get('@x')),
            y=cls.required_float(data.get('@y')),
            diameter=cls.required_float(data.get('@diameter')),
            visible=True,
            north_direction=cls.required_float(data.get('@northDirection')),
            latitude=cls.required_float(data.get('@latitude')),
            longitude=cls.required_float(data.get('@longitude')),
            time_zone=ZoneInfo(time_zone)
        )



import dataclasses
from javaobj import JavaObject
from .Content import Content
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


@dataclasses.dataclass(frozen=True)
class HomeTexture(ModelBase):
    name: str
    image: Content
    width: float
    height: float
    left_to_right_oriented: bool

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'HomeTexture':
        return cls(
            name=o.name,
            image=asset_manager.get_texture(o.name),
            width=o.width,
            height=o.height,
            left_to_right_oriented=o.left_to_right_oriented
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'HomeTexture':
        raise NotImplementedError
        #return cls(
        #    name=data.get('@name'),
        #    image=None,
        #    width=float(data.get('@width')),
        #    height=float(data.get('@height')),
        #    left_to_right_oriented=False
        #)



import dataclasses
from typing import Optional, List, Tuple
from javaobj import JavaObject
from shapely.geometry.point import Point
from shapely.geometry.polygon import Polygon

from .Color import Color
from .HasLevel import HasLevel
from .Renderable import Renderable
from .TextStyle import TextStyle, Alignment
from .Level import Level
from .ModelBase import ModelBase
from ..AssetManager import AssetManager

DEFAULT_TEXT_STYLE = TextStyle(
    font_size=18.0,
    bold=False,
    italic=False,
    alignment=Alignment.CENTER
)


@dataclasses.dataclass
class Label(ModelBase, Renderable, HasLevel):
    identifier: str
    text: str
    x: float
    y: float
    color: Color
    angle: float
    elevation: float
    style: TextStyle
    outline_color: Optional[Color] = None
    pitch: Optional[float] = None

    level: Optional[Level] = None

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Label':
        return cls(
            identifier=o.id,
            text=o.text,
            x=o.x,
            y=o.y,
            style=TextStyle.from_javaobj(o.style, asset_manager) if o.style else DEFAULT_TEXT_STYLE,
            color=Color.from_argb_interger(o.color) if o.color else Color.from_rgba(0, 0, 0, 255),
            outline_color=Color.from_argb_interger(o.outlineColor) if o.outlineColor else None,
            angle=o.angle,
            pitch=o.pitch,
            elevation=o.elevation,
            level=Level.from_javaobj(o.level, asset_manager) if o.level else None,
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Label':
        text_style = data.get('textStyle')
        color = data.get('@color')
        outline_color = data.get('@outlineColor')
        level_identifier = data.get('@level')
        pitch = data.get('@pitch')

        return cls(
            identifier=data.get('@id'),
            text=data.get('text'),
            x=float(data.get('@x')),
            y=float(data.get('@y')),
            style=TextStyle.from_xml_dict(text_style, asset_manager) if text_style else DEFAULT_TEXT_STYLE,
            color=Color.from_hexa(color) if color else Color.from_rgba(0, 0, 0, 255),
            outline_color=Color.from_hexa(outline_color) if outline_color else None,
            angle=float(data.get('@angle', '0.0')),
            pitch=float(pitch) if pitch else None,
            elevation=float(data.get('@elevation', '0.0')),
            level=Level.from_identifier(level_identifier) if level_identifier else None,
        )

    def is_at_level(self, level: Level) -> bool:
        if self.level == level:
            return True

        if self.level and level:
            label_level_elevation = self.level.elevation
            level_elevation = level.elevation
            at_same_level = label_level_elevation == level_elevation and self.level.elevation_index < level.elevation_index
            in_same_level = label_level_elevation < level_elevation < label_level_elevation + self.elevation
            return at_same_level or in_same_level

        return False

    @property
    def points(self) -> List[Tuple[float, float]]:
        return [(self.x, self.y)]

    @property
    def geometry(self) -> Polygon:
        return Point(self.x, self.y).buffer(self.style.font_size)



from typing import List, Tuple, Any
from javaobj import JavaObject
from .HomeMaterial import HomeMaterial
from .HomePieceOfFurniture import HomePieceOfFurniture
from .HomeTexture import HomeTexture
from .Level import Level
from .Sash import Sash
from .TextStyle import TextStyle
from .TextureImage import TextureImage
from ..AssetManager import AssetManager

DEFAULT_CUT_OUT_SHAPE: str = 'M0,0 v1 h1 v-1 z'


class HomeDoorOrWindow(HomePieceOfFurniture):
    sashes: List[Sash]
    wall_thickness: float
    width_depth_deformable: bool
    bound_to_wall: bool
    cut_out_shape: str
    wall_cut_out_on_both_sides: bool
    wall_distance: float
    wall_width: float
    wall_left: float
    wall_height: float
    wall_top: float

    @property
    def points(self) -> List[Tuple[float, float]]:
        raise NotImplementedError

    def __init__(
        self,
        sashes: List[Sash],
        cut_out_shape: str,
        *args: Any,
        wall_thickness: float = 0.0,
        width_depth_deformable: bool = False,
        bound_to_wall: bool = False,
        wall_cut_out_on_both_sides: bool = False,
        wall_distance: float = 0.0,
        wall_width: float = 1.0,
        wall_left: float = 0.0,
        wall_height: float = 1.0,
        wall_top: float = 0.0,
       **kwargs: Any
    ):
        self.sashes = sashes
        self.wall_thickness = wall_thickness
        self.width_depth_deformable = width_depth_deformable
        self.bound_to_wall = bound_to_wall
        self.cut_out_shape = cut_out_shape
        self.wall_cut_out_on_both_sides = wall_cut_out_on_both_sides
        self.wall_distance = wall_distance
        self.wall_width = wall_width
        self.wall_left = wall_left
        self.wall_height = wall_height
        self.wall_top = wall_top
        super().__init__(*args, **kwargs)

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'HomeDoorOrWindow':
        door_or_window = cls(
            # HomeDoorOrWindow
            sashes=[], # Is set later after class init
            wall_thickness=o.wallThickness,
            width_depth_deformable=o.widthDepthDeformable,
            bound_to_wall=o.boundToWall,
            cut_out_shape=o.cutOutShape if o.cutOutShape else DEFAULT_CUT_OUT_SHAPE,
            wall_cut_out_on_both_sides=o.wallCutOutOnBothSides,
            wall_distance=o.wallDistance,
            wall_width=o.wallWidth,
            wall_left=o.wallLeft,
            wall_height=o.wallHeight,
            wall_top=o.wallTop,

            # HomePieceOfFurniture
            identifier=o.id,
            catalog_id=o.catalogId,
            name=o.name,
            name_visible=o.nameVisible,
            name_x_offset=o.nameXOffset,
            name_y_offset=o.nameYOffset,
            name_style=TextStyle.from_javaobj(o.nameStyle, asset_manager) if o.nameStyle else None,
            name_angle=o.nameAngle,
            description=o.description,
            height_in_plan=o.heightInPlan,
            icon=TextureImage.from_content(asset_manager.get_asset_by_uri(o.icon.url.file)) if o.icon else None,
            plan_icon=TextureImage.from_content(asset_manager.get_asset_by_uri(o.planIcon.url.file)) if o.planIcon else None,
            model=asset_manager.get_asset_by_uri(o.model.url.file) if o.model else None,
            width=o.width,
            depth=o.depth,
            height=o.height,
            elevation=o.elevation,
            movable=o.movable,
            is_door_or_window=o.doorOrWindow,
            model_materials=[HomeMaterial.from_javaobj(hmo, asset_manager) for hmo in
                             o.modelMaterials] if o.modelMaterials else None,
            color=o.color,
            texture=HomeTexture.from_javaobj(o.texture, asset_manager) if o.texture else None,
            shininess=o.shininess,
            model_rotation=o.modelRotation,
            staircase_cut_out_shape=o.staircaseCutOutShape,
            back_face_shown=o.backFaceShown,
            resizable=o.resizable,
            deformable=o.deformable,
            texturable=o.texturable,
            price=o.price,
            value_added_tax_percentage=o.valueAddedTaxPercentage,
            currency=o.currency,
            is_visible=o.visible,
            x=o.x,
            y=o.y,
            angle=o.angle,
            is_model_mirrored=o.modelMirrored,
            level=Level.from_javaobj(o.level, asset_manager) if o.level else None
        )

        door_or_window.sashes = [Sash.from_javaobj(door_or_window, sash, asset_manager) for sash in o.sashes]

        return door_or_window

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'HomeDoorOrWindow':
        icon = data.get('@icon')
        icon_content = asset_manager.get_asset(icon) if icon else None

        model = data.get('@model')
        level_identifier = data.get('@level')
        sash = data.get('sash', [])

        if not isinstance(sash, list):
            sash = [sash]

        door_or_window = cls(
            # HomeDoorOrWindow
            sashes=[],  # Is set later after class init
            wall_thickness=cls.required_float(data.get('@wallThickness')),
            width_depth_deformable=data.get('@widthDepthDeformable') == 'true',
            bound_to_wall=data.get('@boundToWall') == 'true',
            cut_out_shape=data.get('@cutOutShape', DEFAULT_CUT_OUT_SHAPE),
            wall_cut_out_on_both_sides=data.get('@wallCutOutOnBothSides') == 'true',
            wall_distance=cls.required_float(data.get('@wallDistance', '0.0')),
            wall_width=cls.required_float(data.get('@wallWidth', '1.0')),
            wall_left=cls.required_float(data.get('@wallLeft', '0.0')),
            wall_height=cls.required_float(data.get('@wallHeight', '1.0')),
            wall_top=cls.required_float(data.get('@wallTop', '0.0')),

            # HomePieceOfFurniture
            identifier=data.get('@id'),
            catalog_id=data.get('@catalogId'),
            name=data.get('@name'),
            name_visible=True,
            name_x_offset=0.0,
            name_y_offset=0.0,
            name_style=None,
            name_angle=0.0,
            description=None,
            height_in_plan=cls.required_float(data.get('@heightInPlan', '-infinity')),
            icon=TextureImage.from_content(icon_content) if icon_content else None,
            plan_icon=None,
            model=asset_manager.get_asset(model) if model else None,
            width=cls.required_float(data.get('@width')),
            depth=cls.required_float(data.get('@depth')),
            height=cls.required_float(data.get('@height')),
            elevation=0.0,
            movable=False,
            is_door_or_window=False,
            model_materials=None,
            color=None,
            texture=None,
            shininess=0.0,
            model_rotation=None,
            staircase_cut_out_shape=None,
            back_face_shown=False,
            resizable=False,
            deformable=False,
            texturable=False,
            price=None,
            value_added_tax_percentage=None,
            currency=None,
            is_visible=True,
            x=cls.required_float(data.get('@x')),
            y=cls.required_float(data.get('@y')),
            angle=cls.required_float(data.get('@angle', '0.0')),
            is_model_mirrored=False,
            level=Level.from_identifier(level_identifier) if level_identifier else None
        )

        door_or_window.sashes = [Sash.from_xml_dict(door_or_window, sash, asset_manager) for sash in sash]

        return door_or_window



import dataclasses
import decimal
import math
from functools import cached_property
from typing import Optional, List, Tuple
from javaobj import JavaObject
from shapely.geometry import Polygon

from .Color import Color
from .HasLevel import HasLevel
from .Level import Level
from .Content import Content
from .Renderable import Renderable
from .TextStyle import TextStyle
from .HomeMaterial import HomeMaterial
from .HomeTexture import HomeTexture
from .ModelBase import ModelBase
from .TextureImage import TextureImage
from ..AssetManager import AssetManager

@dataclasses.dataclass
class HomePieceOfFurniture(ModelBase, Renderable, HasLevel):
    identifier: str
    name_visible: bool
    name_x_offset: float
    name_y_offset: float
    name_angle: float
    width: float
    depth: float
    height: float
    elevation: float
    movable: bool
    is_door_or_window: bool
    back_face_shown: bool
    resizable: bool
    deformable: bool
    texturable: bool
    is_visible: bool
    x: float
    y: float
    angle: float
    is_model_mirrored: bool
    height_in_plan: float
    name_style: Optional[TextStyle] = None
    description: Optional[str] = None
    icon: Optional[TextureImage] = None
    plan_icon: Optional[TextureImage] = None
    model: Optional[Content] = None
    model_materials: Optional[List[HomeMaterial]] = None
    color: Optional[Color] = None
    texture: Optional[HomeTexture] = None
    shininess: Optional[float] = None
    model_rotation: Optional[List[List[float]]] = None
    staircase_cut_out_shape: Optional[str] = None
    price: Optional[decimal.Decimal] = None
    value_added_tax_percentage: Optional[decimal.Decimal] = None
    currency: Optional[str] = None
    catalog_id: Optional[str] = None
    name: Optional[str] = None
    level: Optional[Level] = None

    @cached_property
    def geometry(self) -> Polygon:
        svg_x = self.x - (self.width / 2)
        svg_y = self.y - (self.depth / 2)

        height = self.depth

        corners = [
            (svg_x, svg_y),  # top-left
            (svg_x + self.width, svg_y),  # top-right
            (svg_x + self.width, svg_y + height),  # bottom-right
            (svg_x, svg_y + height)  # bottom-left
        ]

        def rotate(px, py):
            dx = px - self.x
            dy = py - self.y
            cos_a = math.cos(self.angle)
            sin_a = math.sin(self.angle)
            return (
                self.x + dx * cos_a - dy * sin_a,
                self.y + dx * sin_a + dy * cos_a
            )

        rotated_corners = [rotate(px, py) for (px, py) in corners]
        return Polygon(rotated_corners)

    @property
    def points(self) -> List[Tuple[float, float]]:
        return list(self.geometry.exterior.coords)

    def is_at_level(self, level: Level) -> bool:
        if self.level == level:
            return True

        if self.level and level:
            piece_level_elevation = self.level.elevation
            level_elevation = self.elevation
            if piece_level_elevation == level_elevation and self.level.elevation_index < level.elevation_index:
                return True

            if piece_level_elevation < level_elevation and self._is_top_at_level(level):
                return True

        return False

    def _is_top_at_level(self, level: Level) -> bool:
        if not self.level:
            return False
        top_elevation = self.level.elevation + self.elevation + self.height_in_plan
        if self.staircase_cut_out_shape:
            return top_elevation >= level.elevation

        return top_elevation > level.elevation


    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'HomePieceOfFurniture':
        icon_content = asset_manager.get_asset_by_uri(o.icon.url.file) if o.icon else None
        plan_icon_content = asset_manager.get_asset_by_uri(o.planIcon.url.file) if o.planIcon else None

        return cls(
            identifier=o.id,
            catalog_id=o.catalogId,
            name=o.name,
            name_visible=o.nameVisible,
            name_x_offset=o.nameXOffset,
            name_y_offset=o.nameYOffset,
            name_style=TextStyle.from_javaobj(o.nameStyle, asset_manager) if o.nameStyle else None,
            name_angle=o.nameAngle,
            description=o.description,
            height_in_plan=o.heightInPlan,
            icon=TextureImage.from_content(icon_content) if icon_content else None,
            plan_icon=TextureImage.from_content(plan_icon_content) if plan_icon_content else None,
            model=asset_manager.get_asset_by_uri(o.model.url.file) if o.model else None,
            width=o.width,
            depth=o.depth,
            height=o.height,
            elevation=o.elevation,
            movable=o.movable,
            is_door_or_window=o.doorOrWindow,
            model_materials=[HomeMaterial.from_javaobj(hmo, asset_manager) for hmo in o.modelMaterials] if o.modelMaterials else None,
            color=Color.from_argb_interger(o.color) if o.color else None,
            texture=HomeTexture.from_javaobj(o.texture, asset_manager) if o.texture else None,
            shininess=o.shininess,
            model_rotation=o.modelRotation,
            staircase_cut_out_shape=o.staircaseCutOutShape,
            back_face_shown=o.backFaceShown,
            resizable=o.resizable,
            deformable=o.deformable,
            texturable=o.texturable,
            price=o.price,
            value_added_tax_percentage=o.valueAddedTaxPercentage,
            currency=o.currency,
            is_visible=o.visible,
            x=o.x,
            y=o.y,
            angle=o.angle,
            is_model_mirrored=o.modelMirrored,
            level=Level.from_javaobj(o.level, asset_manager) if o.level else None
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'HomePieceOfFurniture':
        icon = data.get('@icon')
        model = data.get('@model')
        level_identifier = data.get('@level')

        return cls(
            identifier=cls.required_str(data.get('@id')),
            catalog_id=data.get('@catalogId'),
            name=data.get('@name'),
            name_visible=True,
            name_x_offset=0.0,
            name_y_offset=0.0,
            name_style=None,
            name_angle=0.0,
            description=None,
            height_in_plan=cls.required_float(data.get('@heightInPlan', '-infinity')),
            icon=TextureImage.from_content(asset_manager.get_asset(icon)) if icon else None,
            plan_icon=None,
            model=asset_manager.get_asset(model) if model else None,
            width=cls.required_float(data.get('@width')),
            depth=cls.required_float(data.get('@depth')),
            height=cls.required_float(data.get('@height')),
            elevation=0.0,
            movable=False,
            is_door_or_window=False,
            model_materials=None,
            color=None,
            texture=None,
            shininess=0.0,
            model_rotation=None,
            staircase_cut_out_shape=None,
            back_face_shown=False,
            resizable=False,
            deformable=False,
            texturable=False,
            price=None,
            value_added_tax_percentage=None,
            currency=None,
            is_visible=True,
            x=cls.required_float(data.get('@x')),
            y=cls.required_float(data.get('@y')),
            angle=cls.required_float(data.get('@angle', '0.0')),
            is_model_mirrored=False,
            level=Level.from_identifier(level_identifier) if level_identifier else None
        )



import math
from dataclasses import dataclass
from typing import Optional, List, Tuple

from javaobj import JavaObject
from shapely.geometry.polygon import Polygon
from shapely.geometry.point import Point

from .Color import Color
from .HasLevel import HasLevel
from .Renderable import Renderable
from .TextStyle import TextStyle, Alignment
from .Level import Level
from .ModelBase import ModelBase
from ..AssetManager import AssetManager

DIMENSION_LINE_MARK_END = (
    'M -5 5',
    'L 5 -5',
    'M 0 5',
    'L 0 -5'
)

VERTICAL_DIMENSION_LINE_DISC = Point(0, 0).buffer(1.5)
VERTICAL_DIMENSION_LINE = Point(0, 0).buffer(5)
DEFAULT_TEXT_STYLE = TextStyle(18, False, False, Alignment.CENTER)

@dataclass
class DimensionLine(ModelBase, Renderable, HasLevel):
    identifier: str
    x_start: float
    y_start: float
    elevation_start: float
    x_end: float
    y_end: float
    elevation_end: float
    offset: float
    end_mark_size: float
    pitch: float
    length_style: TextStyle
    color: Color
    is_visible_in_3d: bool
    level: Optional[Level] = None

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'DimensionLine':
        return cls(
            identifier=o.id,
            x_start=o.xStart,
            y_start=o.yStart,
            elevation_start=o.elevationStart,
            x_end=o.xEnd,
            y_end=o.yEnd,
            elevation_end=o.elevationEnd,
            offset=o.offset,
            end_mark_size=o.endMarkSize,
            pitch=o.pitch,
            color=Color.from_argb_interger(int(o.color)) if o.color else Color.from_rgba(0,0,0,255),
            is_visible_in_3d=o.visibleIn3D,
            length_style=TextStyle.from_javaobj(o.lengthStyle, asset_manager) if o.lengthStyle else DEFAULT_TEXT_STYLE,
            level=Level.from_javaobj(o.level, asset_manager) if o.level else None
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'DimensionLine':
        text_style = data.get('textStyle')
        level_identifier = data.get('@level')
        elevation_end = data.get('@elevationEnd', '0.0')
        elevation_start = data.get('@elevationStart', '0.0')

        color = data.get('@color')
        return cls(
            identifier=data.get('@id'),
            x_start=cls.required_float(data.get('@xStart')),
            y_start=cls.required_float(data.get('@yStart')),
            elevation_start=cls.required_float(elevation_start),
            x_end=cls.required_float(data.get('@xEnd')),
            y_end=cls.required_float(data.get('@yEnd')),
            elevation_end=cls.required_float(elevation_end),
            offset=cls.required_float(data.get('@offset')),
            end_mark_size=cls.required_float(data.get('@endMarkSize', '10.0')),
            pitch=cls.required_float(data.get('@pitch', '0.0')),
            color=Color.from_hexa(color) if color else Color.from_rgba(0,0,0,255),
            is_visible_in_3d=data.get('@visibleIn3D', 'false') == 'true',
            length_style=TextStyle.from_xml_dict(text_style, asset_manager) if text_style else DEFAULT_TEXT_STYLE,
            level=Level.from_identifier(level_identifier) if level_identifier else None
        )


    @property
    def is_elevation_dimension_line(self) -> bool:
        return self.x_start == self.x_end and self.y_start == self.y_end and self.elevation_start != self.elevation_end

    @property
    def length(self) -> float:
        delta_x = self.x_end - self.x_start
        delta_y = self.y_end - self.y_start
        delta_elevation = self.elevation_end - self.elevation_start
        return math.sqrt(delta_x * delta_x + delta_y * delta_y + delta_elevation * delta_elevation)

    @property
    def points(self) -> List[Tuple[float, float]]:
        angle = self.pitch if self.is_elevation_dimension_line else math.atan2(self.y_end - self.y_start, self.x_end - self.x_start)
        dx = -math.sin(angle) * self.offset
        dy = math.cos(angle) * self.offset
        return [
            (self.x_start, self.y_start),
            (self.x_start + dx, self.y_start + dy),
            (self.x_end + dx, self.y_end + dy),
            (self.x_end, self.y_end)
        ]

    @property
    def geometry(self) -> Polygon:
        return Polygon(self.points)

    def is_at_level(self, level: Level) -> bool:
        if self.level == level:
            return True

        if self.level and level:
            dimension_line_level_elevation = self.level.elevation
            level_elevation = level.elevation
            at_same_level = dimension_line_level_elevation == level_elevation and self.level.elevation_index < level.elevation_index
            in_same_level = dimension_line_level_elevation < level_elevation < dimension_line_level_elevation + max(self.elevation_start, self.elevation_end)
            return at_same_level or in_same_level

        return False



import dataclasses
import math
from functools import cached_property
from typing import Optional, Dict, List

from javaobj import JavaObject

from shapely.geometry.polygon import Polygon
from .Baseboard import BaseBoard
from .HasLevel import HasLevel
from .HomeTexture import HomeTexture
from .Level import Level
from .Renderable import Renderable, PointType, PointTypeEditable
from .TextureImage import TextureImage
from .ModelBase import ModelBase
from ..AssetManager import AssetManager

_wall_cache: Dict[str, 'Wall'] = {}



@dataclasses.dataclass
class Wall(ModelBase, Renderable, HasLevel):
    identifier: str
    x_start: float
    y_start: float
    x_end: float
    y_end: float
    thickness: float
    left_side_shininess: float = 0.0
    right_side_shininess: float = 0.0
    symmetric: bool = True
    pattern: Optional[TextureImage] = None
    top_color: Optional[int] = None
    arc_extent: Optional[float] = None
    height: Optional[float] = None
    height_at_end: Optional[float] = None
    left_side_color: Optional[int] = None
    left_side_texture: Optional[HomeTexture] = None
    right_side_color: Optional[int] = None
    right_side_texture: Optional[HomeTexture] = None
    wall_at_start: Optional['Wall'] = None
    wall_at_end: Optional['Wall'] = None
    level: Optional[Level] = None
    left_side_baseboard: Optional[BaseBoard] = None
    right_side_baseboard: Optional[BaseBoard] = None

    _arc_circle_center_cache: Optional[PointTypeEditable] = None
    _points_cache: Optional[List[PointTypeEditable]] = None
    _points_including_baseboards_cache: Optional[List[PointTypeEditable]] = None

    def is_at_level(self, level: Level) -> bool:
        if self.level == level:
            return True
        if self.level and level:
            this_level_elevation = self.level.elevation
            level_elevation = level.elevation
            at_elevation = this_level_elevation == level_elevation and self.level.elevation_index < level.elevation_index
            at_level = this_level_elevation < level_elevation < this_level_elevation + self.get_wall_maximum_height()
            return at_elevation or at_level

        return False

    @cached_property
    def geometry(self) -> Polygon:
        return Polygon(self.points)

    def get_wall_maximum_height(self) -> float:
        if not self.height:
            return 0.0
        if self.height and self.height_at_end:
            return max(self.height, self.height_at_end)

        return self.height

    def is_trapezoidal(self) -> bool:
        if not self.height or not self.height_at_end:
            return False

        return self.height != self.height_at_end

    @property
    def points(self) -> List[PointType]:
        return self.get_points(include_baseboards=False)

    @points.setter
    def points(self, points: List[PointType]) -> None:
        raise NotImplementedError

    def get_points(self, include_baseboards: bool) -> List[PointType]:
        if include_baseboards and (self.left_side_baseboard or self.right_side_baseboard):
            if not self._points_including_baseboards_cache:
                self._points_including_baseboards_cache = self._get_shape_points(include_baseboards)
            return self._points_including_baseboards_cache

        if not self._points_cache:
            self._points_cache = self._get_shape_points(include_baseboards=False)
        return self._points_cache

    def get_arc_circle_center(self) -> PointType:
        if not self.arc_extent:
            raise ValueError('arc_extent is not set')
        if self._arc_circle_center_cache:
            return self._arc_circle_center_cache
        distance = math.dist([self.x_start, self.y_start], [self.x_end, self.y_end])
        if abs(self.arc_extent) > math.pi:
            angle_to_center = -(math.pi + self.arc_extent) / 2
        else:
            angle_to_center = (math.pi - self.arc_extent) / 2

        center_offset = -math.tan(angle_to_center) * distance / 2

        x_mid = (self.x_start + self.x_end) / 2
        y_mid = (self.y_start + self.y_end) / 2

        base_angle = math.atan2(self.x_start - self.x_end, self.y_end - self.y_start)
        self._arc_circle_center_cache = [x_mid + center_offset * math.cos(base_angle), y_mid + center_offset * math.sin(base_angle)]

        return self._arc_circle_center_cache

    def _compute_round_wall_shape_point(
            self,
            wall_points: List[PointType],
            angle: float,
            index: int,
            angle_delta: float,
            arc_circle_center: PointType,
            exterior_arc_radius: float,
            interior_arc_radius: float
    ) -> None:
        cos = math.cos(angle)
        sin = math.sin(angle)

        interior_arc_point = [
            arc_circle_center[0] + interior_arc_radius * cos,
            arc_circle_center[1] - interior_arc_radius * sin
        ]


        exterior_arc_point = [
            arc_circle_center[0] + exterior_arc_radius * cos,
            arc_circle_center[1] - exterior_arc_radius * sin
        ]

        if angle_delta > 0:
            wall_points.insert(index, interior_arc_point)
            wall_points.insert(len(wall_points) - index, exterior_arc_point)
        else:
            wall_points.insert(index, exterior_arc_point)
            wall_points.insert(len(wall_points) - index, interior_arc_point)

    def _get_unjoined_shape_points(self, include_baseboards: bool) -> List[PointType]:
        epsilon = 1e-10

        if (
                self.arc_extent and
                self.arc_extent != 0 and
                (self.x_start - self.x_end) ** 2 + (self.y_start - self.y_end) ** 2 > epsilon
        ):

            arc_center = self.get_arc_circle_center()
            start_angle = math.atan2(arc_center[1] - self.y_start, arc_center[0] - self.x_start)
            start_angle += 2 * math.atan2(self.y_start - self.y_end, self.x_end - self.x_start)

            arc_radius = math.dist(arc_center, [self.x_start, self.y_start])
            exterior_radius = arc_radius + self.thickness / 2
            interior_radius = max(0.0, arc_radius - self.thickness / 2)
            arc_length = exterior_radius * abs(self.arc_extent)
            angle_delta = self.arc_extent / math.sqrt(arc_length) if arc_length != 0 else 0
            angle_steps = int(self.arc_extent / angle_delta) if angle_delta != 0 else 0

            if include_baseboards:
                if angle_delta > 0:
                    if self.left_side_baseboard:
                        exterior_radius += self.left_side_baseboard.thickness
                    if self.right_side_baseboard:
                        interior_radius -= self.right_side_baseboard.thickness
                else:
                    if self.left_side_baseboard:
                        interior_radius -= self.left_side_baseboard.thickness
                    if self.right_side_baseboard:
                        exterior_radius += self.right_side_baseboard.thickness

            wall_points: List[PointType] = []

            if self.symmetric:
                if abs(self.arc_extent - angle_steps * angle_delta) > 1e-6:
                    angle_steps += 1
                    angle_delta = self.arc_extent / angle_steps
                for i in range(angle_steps + 1):
                    angle = start_angle + self.arc_extent - i * angle_delta
                    self._compute_round_wall_shape_point(
                        wall_points, angle, i, angle_delta,
                        arc_center, exterior_radius, interior_radius)
            else:
                i = 0
                angle = self.arc_extent

                def _step_condition(a: float) -> bool:
                    return a >= angle_delta * 0.1 if angle_delta > 0 else a <= -angle_delta * 0.1

                while _step_condition(angle):
                    self._compute_round_wall_shape_point(
                        wall_points, start_angle + angle, i, angle_delta,
                        arc_center, exterior_radius, interior_radius)
                    angle -= angle_delta
                    i += 1
                self._compute_round_wall_shape_point(
                    wall_points, start_angle, i, angle_delta,
                    arc_center, exterior_radius, interior_radius)

            return wall_points

        angle = math.atan2(self.y_end - self.y_start, self.x_end - self.x_start)
        sin_a = math.sin(angle)
        cos_a = math.cos(angle)

        left_thickness = self.thickness / 2
        if include_baseboards and self.left_side_baseboard:
            left_thickness += self.left_side_baseboard.thickness
        left_dx = sin_a * left_thickness
        left_dy = cos_a * left_thickness

        right_thickness = self.thickness / 2
        if include_baseboards and self.right_side_baseboard:
            right_thickness += self.right_side_baseboard.thickness
        right_dx = sin_a * right_thickness
        right_dy = cos_a * right_thickness

        return [
            [self.x_start + left_dx, self.y_start - left_dy],
            [self.x_end + left_dx, self.y_end - left_dy],
            [self.x_end - right_dx, self.y_end + right_dy],
            [self.x_start - right_dx, self.y_start + right_dy]
        ]

    def _compute_intersection(
            self,
            point1: PointTypeEditable,
            point2: PointTypeEditable,
            point3: PointTypeEditable,
            point4: PointTypeEditable,
            limit: float
    ) -> None:
        try:
            alpha1 = (point2[1] - point1[1]) / (point2[0] - point1[0])
        except ZeroDivisionError:
            alpha1 = float('inf') if point2[1] >= point1[1] else -float('inf')

        try:
            alpha2 = (point4[1] - point3[1]) / (point4[0] - point3[0])
        except ZeroDivisionError:
            alpha2 = float('inf') if point4[1] >= point3[1] else -float('inf')

        if alpha1 != alpha2:
            x = point1[0]
            y = point1[1]

            # First line vertical
            if math.isinf(alpha1) and not math.isinf(alpha2):
                x = point1[0]
                beta2 = point4[1] - alpha2 * point4[0]
                y = alpha2 * x + beta2

            # Second line vertical
            elif math.isinf(alpha2) and not math.isinf(alpha1):
                x = point3[0]
                beta1 = point2[1] - alpha1 * point2[0]
                y = alpha1 * x + beta1

            # Neither vertical
            elif not math.isinf(alpha1) and not math.isinf(alpha2):
                same_signum = math.copysign(1, alpha1) == math.copysign(1, alpha2)
                if (abs(alpha1 - alpha2) > 1E-5 and
                        (not same_signum or (max(abs(alpha1), abs(alpha2)) / min(abs(alpha1), abs(alpha2))) > 1.004)):
                    beta1 = point2[1] - alpha1 * point2[0]
                    beta2 = point4[1] - alpha2 * point4[0]
                    x = (beta2 - beta1) / (alpha1 - alpha2)
                    y = alpha1 * x + beta1

            # Only update point1 if the new intersection is within the limit
            if (point1[0] - x) ** 2 + (point1[1] - y) ** 2 < limit * limit:
                point1[0] = x
                point1[1] = y

    def _get_shape_points(self, include_baseboards: bool) -> List[PointType]:
        epsilon = 0.01
        wall_points = self._get_unjoined_shape_points(include_baseboards)
        left_start = 0
        wall_points_len = len(wall_points)
        right_start = wall_points_len - 1
        left_end = wall_points_len // 2 - 1
        right_end = wall_points_len// 2

        if self.wall_at_start:
            wall_start = self.wall_at_start
            start_points = wall_start._get_unjoined_shape_points(include_baseboards)
            start_left_start = 0
            start_right_start = len(start_points) - 1
            start_left_end = len(start_points) // 2 - 1
            start_right_end = len(start_points) // 2

            joined_at_end = (
                    wall_start.wall_at_end == self and
                    (wall_start.wall_at_start != self or
                     (wall_start.x_end == self.x_start and wall_start.y_end == self.y_start))
            )
            joined_at_start = (
                    wall_start.wall_at_start == self and
                    (wall_start.wall_at_end != self or
                     (wall_start.x_start == self.x_start and wall_start.y_start == self.y_start))
            )

            start_cache = wall_start._points_including_baseboards_cache if include_baseboards else wall_start._points_cache

            limit = 2 * max(self.thickness, wall_start.thickness)

            if joined_at_end:
                self._compute_intersection(wall_points[left_start], wall_points[left_start + 1],
                                          start_points[start_left_end], start_points[start_left_end - 1], limit)
                self._compute_intersection(wall_points[right_start], wall_points[right_start - 1],
                                          start_points[start_right_end], start_points[start_right_end + 1], limit)

                if start_cache:
                    if (abs(wall_points[left_start][0] - start_cache[start_left_end][0]) < epsilon and
                            abs(wall_points[left_start][1] - start_cache[start_left_end][1]) < epsilon):
                        wall_points[left_start] = start_cache[start_left_end]
                    if (abs(wall_points[right_start][0] - start_cache[start_right_end][0]) < epsilon and
                            abs(wall_points[right_start][1] - start_cache[start_right_end][1]) < epsilon):
                        wall_points[right_start] = start_cache[start_right_end]

            elif joined_at_start:
                self._compute_intersection(wall_points[left_start], wall_points[left_start + 1],
                                          start_points[start_right_start], start_points[start_right_start - 1], limit)
                self._compute_intersection(wall_points[right_start], wall_points[right_start - 1],
                                          start_points[start_left_start], start_points[start_left_start + 1], limit)

                if start_cache:
                    if (abs(wall_points[left_start][0] - start_cache[start_right_start][0]) < epsilon and
                            abs(wall_points[left_start][1] - start_cache[start_right_start][1]) < epsilon):
                        wall_points[left_start] = start_cache[start_right_start]
                    if (abs(wall_points[right_start][0] - start_cache[start_left_start][0]) < epsilon and
                            abs(wall_points[right_start][1] - start_cache[start_left_start][1]) < epsilon):
                        wall_points[right_start] = start_cache[start_left_start]

        if self.wall_at_end:
            wall_end = self.wall_at_end
            end_points = wall_end._get_unjoined_shape_points(include_baseboards)
            end_left_start = 0
            end_right_start = len(end_points) - 1
            end_left_end = len(end_points) // 2 - 1
            end_right_end = len(end_points) // 2

            joined_at_start = (
                    wall_end.wall_at_start == self and
                    (wall_end.wall_at_end != self or
                     (wall_end.x_start == self.x_end and wall_end.y_start == self.y_end))
            )

            joined_at_end = (
                    wall_end.wall_at_end == self and
                    (wall_end.wall_at_start != self or
                     (wall_end.x_end == self.x_end and wall_end.y_end == self.y_end))
            )

            end_cache = wall_end._points_including_baseboards_cache if include_baseboards else wall_end._points_cache

            limit = 2 * max(self.thickness, wall_end.thickness)

            if joined_at_start:
                self._compute_intersection(wall_points[left_end], wall_points[left_end - 1],
                                          end_points[end_left_start], end_points[end_left_start + 1], limit)
                self._compute_intersection(wall_points[right_end], wall_points[right_end + 1],
                                          end_points[end_right_start], end_points[end_right_start - 1], limit)

                if end_cache:
                    if (abs(wall_points[left_end][0] - end_cache[end_left_start][0]) < epsilon and
                            abs(wall_points[left_end][1] - end_cache[end_left_start][1]) < epsilon):
                        wall_points[left_end] = end_cache[end_left_start]
                    if (abs(wall_points[right_end][0] - end_cache[end_right_start][0]) < epsilon and
                            abs(wall_points[right_end][1] - end_cache[end_right_start][1]) < epsilon):
                        wall_points[right_end] = end_cache[end_right_start]

            elif joined_at_end:
                self._compute_intersection(wall_points[left_end], wall_points[left_end - 1],
                                          end_points[end_right_end], end_points[end_right_end + 1], limit)
                self._compute_intersection(wall_points[right_end], wall_points[right_end + 1],
                                          end_points[end_left_end], end_points[end_left_end - 1], limit)

                if end_cache:
                    if (abs(wall_points[left_end][0] - end_cache[end_right_end][0]) < epsilon and
                            abs(wall_points[left_end][1] - end_cache[end_right_end][1]) < epsilon):
                        wall_points[left_end] = end_cache[end_right_end]
                    if (abs(wall_points[right_end][0] - end_cache[end_left_end][0]) < epsilon and
                            abs(wall_points[right_end][1] - end_cache[end_left_end][1]) < epsilon):
                        wall_points[right_end] = end_cache[end_left_end]


        return wall_points

    @classmethod
    def from_identifier(cls, identifier: str) -> 'Wall':
        wall = _wall_cache.get(identifier)
        if not wall:
            raise ValueError('Identifier not found in cache')
        return wall

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Wall':
        wall_identifier = o.id

        if wall_identifier in _wall_cache:
            return _wall_cache[wall_identifier]

        # First pass: create partially initialized object
        wall = cls(
            identifier=wall_identifier,
            x_start=o.xStart,
            y_start=o.yStart,
            x_end=o.xEnd,
            y_end=o.yEnd,
            thickness=getattr(o, 'thickness', 0.0)
        )

        pattern = getattr(o, 'pattern', None)
        height = getattr(o, 'height', None)
        arc_extent = getattr(o, 'arcExtent', None)
        level = getattr(o, 'level', None)

        _wall_cache[wall_identifier] = wall
        # Second pass: populate fields
        wall.arc_extent = arc_extent.value if arc_extent else None
        wall.height = height.value if height else None
        wall.height_at_end = getattr(o, 'heightAtEnd', None)
        wall.left_side_color = getattr(o, 'leftSideColor', None)
        wall.left_side_texture = getattr(o, 'leftSideTexture', None)
        wall.left_side_shininess = getattr(o, 'leftSideShininess', 0.0)
        wall.left_side_baseboard = getattr(o, 'leftSideBaseboard', None)
        wall.right_side_color = getattr(o, 'rightSideColor', None)
        wall.right_side_texture = getattr(o, 'rightSideTexture', None)
        wall.right_side_shininess = getattr(o, 'rightSideShininess', 0.0)
        wall.right_side_baseboard = getattr(o, 'rightSideBaseboard', None)
        wall.pattern = TextureImage.from_javaobj(pattern, asset_manager) if pattern else None
        wall.top_color = getattr(o, 'topColor', None)
        wall.symmetric = getattr(o, 'symmetric', True)

        wall_at_start = getattr(o, 'wallAtStart', None)
        wall_at_end = getattr(o, 'wallAtEnd', None)

        wall.wall_at_start = cls.from_javaobj(wall_at_start, asset_manager) if wall_at_start else None
        wall.wall_at_end = cls.from_javaobj(wall_at_end, asset_manager) if wall_at_end else None

        wall.level = Level.from_javaobj(level, asset_manager) if level else None

        return wall

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Wall':
        pattern = data.get('@pattern')
        level_identifier = data.get('@level')
        wall_identifier = cls.required_str(data.get('@id'))

        if wall_identifier in _wall_cache:
            return _wall_cache[wall_identifier]

        # First pass: create partially initialized object
        wall = cls(
            identifier=wall_identifier,
            x_start=cls.required_float(data.get('@xStart')),
            y_start=cls.required_float(data.get('@yStart')),
            x_end=cls.required_float(data.get('@xEnd')),
            y_end=cls.required_float(data.get('@yEnd')),
            thickness=float(data.get('@thickness', '0.0'))
        )

        _wall_cache[wall_identifier] = wall

        # Second pass: populate fields
        wall.height = cls.required_float(data.get('@height'))
        wall.pattern = TextureImage.from_str(pattern, asset_manager) if pattern else None

        wall_at_start = data.get('@wallAtStart')
        wall_at_end = data.get('@wallAtEnd')
        wall.wall_at_start = _wall_cache.get(wall_at_start) if wall_at_start else None
        wall.wall_at_end = _wall_cache.get(wall_at_end) if wall_at_end else None

        wall.level = Level.from_identifier(level_identifier) if level_identifier else None

        return wall



from typing import Optional
from abc import ABC, abstractmethod
from sh3d.model.Level import Level


class HasLevel(ABC):

    @property
    @abstractmethod
    def level(self) -> Optional[Level]:
        raise NotImplementedError

    def is_at_level(self, level: Level) -> bool:
        raise NotImplementedError



import enum
from typing import Optional, List, Union
from dataclasses import dataclass
from functools import cached_property
import numpy as np
from javaobj import JavaObject
from shapely.geometry.polygon import Polygon
from shapely.geometry.linestring import LineString

from .Color import Color
from .HasLevel import HasLevel
from .Renderable import Renderable, PointType
from .Level import Level
from .ModelBase import ModelBase
from ..AssetManager import AssetManager

@enum.unique
class CapStyle(enum.Enum):
    BUTT = 'BUTT'
    SQUARE = 'SQUARE'
    ROUND = 'ROUND'


@enum.unique
class JoinStyle(enum.Enum):
    BEVEL = 'BEVEL'
    MITER = 'MITER'
    ROUND = 'ROUND'
    CURVED = 'CURVED'

@enum.unique
class ArrowStyle(enum.Enum):
    NONE = 'NONE'
    DELTA = 'DELTA'
    OPEN = 'OPEN'
    DISC = 'DISC'

@enum.unique
class DashStyle(enum.Enum):
    SOLID = 'SOLID'
    DOT = 'DOT'
    DASH = 'DASH'
    DASH_DOT = 'DASH_DOT'
    DASH_DOT_DOT = 'DASH_DOT_DOT'
    CUSTOMIZED = 'CUSTOMIZED'


    @property
    def dash_pattern(self) -> Optional[List[float]]:
        return {
            DashStyle.SOLID: [1.0, 0.0],
            DashStyle.DOT: [1.0, 1.0],
            DashStyle.DASH: [4.0, 2.0],
            DashStyle.DASH_DOT: [8.0, 2.0, 2.0, 2.0],
            DashStyle.DASH_DOT_DOT: [8.0, 2.0, 2.0, 2.0, 2.0, 2.0],
        }.get(self)




@dataclass
class Polyline(ModelBase, Renderable, HasLevel):
    identifier: str

    thickness: float
    cap_style_name: str
    join_style_name: str
    dash_style_name: str
    dash_pattern: List[float]
    dash_offset: float
    start_arrow_style_name: str
    end_arrow_style_name: str
    is_closed_path: bool
    color: Color
    elevation: float
    points: List[PointType]
    cap_style: CapStyle = CapStyle.BUTT
    join_style: JoinStyle = JoinStyle.MITER
    dash_style: DashStyle = DashStyle.SOLID
    start_arrow_style: ArrowStyle = ArrowStyle.NONE
    end_arrow_style: ArrowStyle = ArrowStyle.NONE
    level: Optional[Level] = None

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'Polyline':
        return cls(
            identifier=o.id,
            points=o.points,
            thickness=o.thickness,
            cap_style=CapStyle(o.capStyleName) if o.capStyleName else CapStyle.BUTT,
            cap_style_name=o.capStyleName,
            join_style=JoinStyle(o.joinStyleName) if o.joinStyleName else JoinStyle.MITER,
            join_style_name=o.joinStyleName,
            dash_style=DashStyle(o.dashStyleName) if o.dashStyleName else DashStyle.SOLID,
            dash_style_name=o.dashStyleName,
            dash_pattern=o.dashPattern if o.dashPattern else DashStyle(o.dashStyleName).dash_pattern if o.dashStyleName else None,
            dash_offset=o.dashOffset,
            start_arrow_style=ArrowStyle(o.startArrowStyleName) if o.startArrowStyleName else None,
            start_arrow_style_name=o.startArrowStyleName,
            end_arrow_style=ArrowStyle(o.endArrowStyleName) if o.endArrowStyleName else None,
            end_arrow_style_name=o.endArrowStyleName,
            is_closed_path=o.closedPath,
            color=Color.from_argb_interger(o.color) if o.color else None,
            elevation=o.elevation,
            level=Level.from_javaobj(o.level, asset_manager) if o.level else None
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'Polyline':
        points = data.get('point')
        cap_style_name = data.get('@capStyle')
        join_style_name = data.get('@joinStyle')
        dash_style_name = data.get('@dashStyle')
        dash_pattern = data.get('@dashPattern')
        start_arrow_style_name = data.get('@startArrowStyle')
        end_arrow_style_name = data.get('@endArrowStyle')
        color = data.get('@color')
        level_identifier = data.get('@level')
        return cls(
            identifier=data.get('@id'),
            points=[[float(point.get('@x')), float(point.get('@y'))] for point in points],
            thickness=cls.required_float(data.get('@thickness')),
            cap_style=CapStyle(cap_style_name) if cap_style_name else CapStyle.BUTT,
            cap_style_name=cap_style_name,
            join_style=JoinStyle(join_style_name) if join_style_name else JoinStyle.MITER,
            join_style_name=join_style_name,
            dash_style=DashStyle(dash_style_name) if dash_style_name else DashStyle.SOLID,
            dash_style_name=dash_style_name,
            dash_pattern=dash_pattern if dash_pattern else DashStyle(dash_style_name).dash_pattern if dash_style_name else None,
            dash_offset=data.get('@dashOffset'),
            start_arrow_style=ArrowStyle(start_arrow_style_name) if start_arrow_style_name else None,
            start_arrow_style_name=start_arrow_style_name,
            end_arrow_style=ArrowStyle(end_arrow_style_name) if end_arrow_style_name else None,
            end_arrow_style_name=end_arrow_style_name,
            is_closed_path=data.get('@closedPath', False),
            color=Color.from_hexa(color) if color else None,
            elevation=data.get('@elevation'),
            level=Level.from_identifier(level_identifier) if level_identifier else None
        )


    @cached_property
    def polyline_path(self) -> Union[LineString | Polygon]:

        points = self.points
        n = len(points)
        path_points = []

        if self.join_style == JoinStyle.CURVED:
            def cubic_bezier(p0: np.ndarray, c1: np.ndarray, c2: np.ndarray, p1: np.ndarray, steps: int =20) -> np.ndarray:
                t = np.linspace(0, 1, steps)
                curve = (1 - t)[:, None] ** 3 * p0 + \
                        3 * (1 - t)[:, None] ** 2 * t[:, None] * c1 + \
                        3 * (1 - t)[:, None] * t[:, None] ** 2 * c2 + \
                        t[:, None] ** 3 * p1
                return curve

            limit = n if self.is_closed_path else n - 1
            for i in range(limit):
                prev_idx = (i - 1) % n
                next_idx = (i + 1) % n
                next_next_idx = (i + 2) % n

                prev = np.array(points[prev_idx])
                curr = np.array(points[i])
                next_pt = np.array(points[next_idx])
                next_next = np.array(points[next_next_idx])

                v1 = (next_pt - prev) / 3.625 if i != 0 or self.is_closed_path else np.zeros(2)
                v2 = (curr - next_next) / 3.625 if i != n - 2 or self.is_closed_path else np.zeros(2)

                control1 = curr + v1
                control2 = next_pt + v2
                bezier = cubic_bezier(curr, control1, control2, next_pt)

                path_points.extend(bezier.tolist())

            if self.is_closed_path:
                path_points.append(path_points[0])  # close it manually
            return LineString(path_points)

        if self.is_closed_path:
            return Polygon(points)

        return LineString(points)

    @property
    def geometry(self) -> Polygon:
        cap_style = {
            CapStyle.ROUND: 'round',
            CapStyle.SQUARE: 'square',
            CapStyle.BUTT: 'flat'
        }.get(self.cap_style, 'round')

        join_style = {
            JoinStyle.ROUND: 'round',
            JoinStyle.MITER: 'mitre',
            JoinStyle.BEVEL: 'bevel'
        }.get(self.join_style, 'round')

        return self.polyline_path.buffer(self.thickness, cap_style=cap_style, join_style=join_style)

    def is_at_level(self, level: Level) -> bool:
        if self.level == level:
            return True

        if self.level and level:
            dimension_line_level_elevation = self.level.elevation
            level_elevation = level.elevation
            return dimension_line_level_elevation == level_elevation and self.level.elevation_index < level.elevation_index

        return False



import dataclasses
import zipfile
import hashlib
from functools import cached_property
from pathlib import Path

from .ContentDigest import ContentDigest
from ..enums.DigestHashEnum import DigestHashEnum


class IDataLoader:

    def load(self) -> bytes:
        raise NotImplementedError

    def digest(self) -> ContentDigest:
        raise NotImplementedError


class PathLoader(IDataLoader):
    path: Path

    def __init__(self, path: Path):
        self.path = path

    def load(self) -> bytes:
        with self.path.open('rb') as f:
            return f.read()

    def digest(self) -> ContentDigest:
        return ContentDigest(self.path.name, DigestHashEnum.SHA_1, hashlib.sha1(self.load(), usedforsecurity=False).digest())


class ZipLoader(IDataLoader):
    enum_to_hash = {
        DigestHashEnum.SHA_1: hashlib.sha1
    }

    def __init__(self, zip_file: zipfile.ZipFile, content_digest: ContentDigest):
        self.zip_file = zip_file
        self.content_digest = content_digest

    def load(self) -> bytes:
        asset = self.zip_file.read(self.content_digest.name)
        calculated_digest = self.enum_to_hash.get(self.content_digest.digest_hash, hashlib.sha1)(asset).digest()
        if calculated_digest != self.content_digest.digest:
            raise ValueError('Incorrect digest')
        return asset

    def digest(self) -> ContentDigest:
        return self.content_digest


@dataclasses.dataclass(frozen=True)
class Content:
    data_loader: IDataLoader


    @cached_property
    def content_digest(self) -> ContentDigest:
        return self.data_loader.digest()

    @cached_property
    def data(self) -> bytes:
        return self.data_loader.load()

    @classmethod
    def from_content_digest(cls, zip_file: zipfile.ZipFile, content_digest: ContentDigest) -> 'Content':
        return cls(
            data_loader=ZipLoader(zip_file, content_digest),
        )

    @classmethod
    def from_path(cls, path: Path) -> 'Content':
        return cls(
            data_loader=PathLoader(path),
        )



import math
import enum
import dataclasses
from typing import Union

from shapely.geometry.linestring import LineString
from shapely.geometry.polygon import Polygon

@enum.unique
class ArcTypeEnum(enum.Enum):
    OPEN = 'OPEN'
    CHORD = 'CHORD'
    PIE = 'PIE'

def arc(
        x: float,
        y: float,
        width: float,
        height: float,
        start_angle_deg: float,
        arc_angle_deg: float,
        arc_type: ArcTypeEnum=ArcTypeEnum.OPEN,
        num_points: int=32
) -> Union[LineString, Polygon]:
    cx = x + width / 2
    cy = y + height / 2
    rx = width / 2
    ry = height / 2

    start_rad = math.radians(start_angle_deg)
    end_rad = math.radians(start_angle_deg + arc_angle_deg)

    # Normalize angle range for clockwise or counter-clockwise
    angle_step = (end_rad - start_rad) / num_points

    points = [
        (
            cx + rx * math.cos(start_rad + i * angle_step),
            cy + ry * math.sin(start_rad + i * angle_step)
        )
        for i in range(num_points + 1)
    ]

    if arc_type == ArcTypeEnum.OPEN:
        return LineString(points)

    if arc_type == ArcTypeEnum.CHORD:
        # Add straight line from end to start to close with a chord
        return LineString(points + [points[0]])

    if arc_type == ArcTypeEnum.PIE:
        # Add center and return as filled sector
        return Polygon([ (cx, cy) ] + points + [ (cx, cy) ])

    raise ValueError(f"Unsupported arc type: {arc_type}")

@dataclasses.dataclass
class Rectangle:
    x: float
    y: float
    width: float
    height: float

    @property
    def max_x(self) -> float:
        return self.x + self.width

    @property
    def max_y(self) -> float:
        return self.y + self.height

    def add(self, new_x: float, new_y: float) -> None:
        # Expand bounds to include point (px, py)
        self.x = min(self.x, new_x)
        self.y =  min(self.y, new_y)
        self.width = max(self.max_x, new_x) - self.x
        self.height = max(self.max_y, new_y) - self.y


    def add_rectangle(self, rect: 'Rectangle') -> None:
        x1 = min(self.x, rect.x)
        x2 = max(self.max_x, rect.max_x)
        y1 = min(self.y, rect.y)
        y2 = max(self.max_y, rect.max_y)

        self.x = x1
        self.y = y1
        self.width = x2 - x1
        self.height =  y2 - y1


    def to_polygon(self) -> Polygon:
        return Polygon([
            (self.x, self.y),
            (self.x + self.width, self.y),
            (self.x + self.width, self.y + self.height),
            (self.x, self.y + self.height),
            (self.x, self.y)  # Close the polygon
        ])



import io
import weakref
import dataclasses
from typing import Any
from PIL import Image
from javaobj import JavaObject
from .Content import Content
from ..AssetManager import AssetManager
from .ModelBase import ModelBase


@dataclasses.dataclass(frozen=True)
class TextureImage(ModelBase):
    _instances = weakref.WeakValueDictionary()
    name: str
    image: Content
    width: float
    height: float

    def __new__(cls, name: str, *_args: Any, **_kwargs: Any) -> 'TextureImage':
        if name in cls._instances:
            return cls._instances[name]
        instance = super().__new__(cls)
        cls._instances[name] = instance
        return instance

    @classmethod
    def from_content(cls, content: Content) -> 'TextureImage':
        image_info = Image.open(io.BytesIO(content.data))
        return cls(
            name=content.content_digest.name,
            image=content,
            width=image_info.width,
            height=image_info.height
        )

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'TextureImage':
        image = asset_manager.get_pattern(o.name)
        image_info = Image.open(io.BytesIO(image.data))

        return cls(
            name=o.name,
            image=image,
            width=image_info.width,
            height=image_info.height
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'ModelBase':
        raise NotImplementedError

    @classmethod
    def from_str(cls, pattern_name: str, asset_manager: AssetManager) -> 'TextureImage':
        image = asset_manager.get_pattern(pattern_name)
        image_info = Image.open(io.BytesIO(image.data))
        return cls(
            name=pattern_name,
            image=image,
            width=image_info.width,
            height=image_info.height
        )



from pathlib import Path
from sh3d.FileLoader import FileLoader, HomeSource
from sh3d.model.Home import Home


def test_can_init(sh3d_path: Path) -> None:
    with FileLoader(sh3d_path) as file_loader:
        assert isinstance(file_loader.home, Home)

def test_can_init_path_str(sh3d_path: Path) -> None:
    with FileLoader(str(sh3d_path)) as file_loader:
        assert isinstance(file_loader.home, Home)

def test_can_init_custom_resources(sh3d_path: Path, resources_path: Path) -> None:
    with FileLoader(sh3d_path, resources_path) as file_loader:
        assert isinstance(file_loader.home, Home)

def test_can_init_custom_resources_str(sh3d_path: Path, resources_path: Path) -> None:
    with FileLoader(sh3d_path, str(resources_path)) as file_loader:
        assert isinstance(file_loader.home, Home)

def test_can_init_xml(sh3d_path: Path) -> None:
    with FileLoader(sh3d_path, home_source=HomeSource.XML) as file_loader:
        assert isinstance(file_loader.home, Home)


from functools import cached_property
from typing import List, Tuple, Union
from shapely.geometry.base import BaseGeometry

PointTypeFixed = Tuple[float, float]
PointTypeEditable = List[float]

PointType = Union[PointTypeFixed, PointTypeEditable]

class Renderable:
    identifier: str
    points: List[PointType]

    @cached_property
    def geometry(self) -> BaseGeometry:
        raise NotImplementedError



import dataclasses
from typing import Optional
from javaobj import JavaObject
from .HomeTexture import HomeTexture
from .ModelBase import ModelBase
from ..AssetManager import AssetManager


@dataclasses.dataclass
class HomeEnvironment(ModelBase):
    observer_camera_elevation_adjusted: bool
    ground_color: int
    ground_texture: Optional['HomeTexture']
    sky_color: int
    sky_texture: Optional['HomeTexture']
    light_color: int
    ceiling_light_color: int
    walls_alpha: float
    subpart_size_under_light: float
    all_levels_visible: bool
    photo_width: int
    photo_height: int
    photo_aspect_ratio_name: Optional[str]
    photo_quality: int
    video_width: int
    video_aspect_ratio_name: Optional[str]
    video_quality: int
    video_frame_rate: int

    @classmethod
    def from_javaobj(cls, o: JavaObject, asset_manager: AssetManager) -> 'HomeEnvironment':
        return cls(
            observer_camera_elevation_adjusted=o.observerCameraElevationAdjusted,
            ground_color=o.groundColor,
            ground_texture=HomeTexture.from_javaobj(o.groundTexture, asset_manager) if o.groundTexture else None,
            sky_color=o.skyColor,
            sky_texture=HomeTexture.from_javaobj(o.skyTexture, asset_manager) if o.skyTexture else None,
            light_color=o.lightColor,
            ceiling_light_color=o.ceilingLightColor,
            walls_alpha=o.wallsAlpha,
            subpart_size_under_light=o.subpartSizeUnderLight,
            all_levels_visible=o.allLevelsVisible,
            photo_width=o.photoWidth,
            photo_height=o.photoHeight,
            photo_aspect_ratio_name=o.photoAspectRatioName,
            photo_quality=o.photoQuality,
            video_width=o.videoWidth,
            video_aspect_ratio_name=o.videoAspectRatioName,
            video_quality=o.videoQuality,
            video_frame_rate=o.videoFrameRate
        )

    @classmethod
    def from_xml_dict(cls, data: dict, asset_manager: AssetManager) -> 'HomeEnvironment':
        ground_texture = data.get('@groundTexture')
        sky_texture = data.get('@skyTexture')

        return cls(
            observer_camera_elevation_adjusted=False,
            ground_color=data.get('@groundColor'),
            ground_texture=HomeTexture.from_xml_dict(ground_texture, asset_manager) if ground_texture else None,
            sky_color=data.get('@skyColor'),
            sky_texture=HomeTexture.from_xml_dict(sky_texture, asset_manager) if sky_texture else None,
            light_color=data.get('@lightColor'),
            ceiling_light_color=data.get('@ceillingLightColor'),
            walls_alpha=0.0,
            subpart_size_under_light=0.0,
            all_levels_visible=True,
            photo_width=cls.required_int(data.get('@photoWidth')),
            photo_height=cls.required_int(data.get('@photoHeight')),
            photo_aspect_ratio_name=data.get('@photoAspectRatio'),
            photo_quality=data.get('@photoQuality'),
            video_width=data.get('@videoWidth'),
            video_aspect_ratio_name=data.get('@videoAspectRatio'),
            video_quality=data.get('@videoQuality'),
            video_frame_rate=data.get('@videoFrameRate')
        )
